#include <iostream>
#include <string>
#include <map>

#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/imgcodecs.hpp>

#include "include/tcon_dev.h"

using namespace cv;

int main(int argc, char *argv[]) {
    /* Maps any T-CON error code to a descriptive string. */
    std::map<short, std::string> error_to_string = {
            { LIBTCON_SUCCESS, "Success" },
            { LIBTCON_ERROR_IO, "Input / Output error" },
            { LIBTCON_ERROR_INVALID_PARAM, "Invalid parameters" },
            { LIBTCON_ERROR_ACCESS, "Access denied (insufficient permission)" },
            { LIBTCON_ERROR_NO_DEVICE, "No such device" },
            { LIBTCON_ERROR_NOT_FOUND, "Entity not found" },
            { LIBTCON_ERROR_BUSY, "Resource busy" },
            { LIBTCON_ERROR_TIMEOUT, "Operation timed out" },
            { LIBTCON_ERROR_OVERFLOW, "Overflow" },
            { LIBTCON_ERROR_PIPE, "Pipe error" },
            { LIBTCON_ERROR_INTERRUPTED, "System call interrupted" },
            { LIBTCON_ERROR_NO_MEM, "Insufficient memory" },
            { LIBTCON_ERROR_NOT_SUPPORTED, "Operation not supported or unimplemented on this platform" },
            { LIBTCON_ERROR_OTHER, "Other error" },
        };

    std::cout << "ejump_usb - Displays an image onto a connected e-paper via an eJump T-CON board" << std::endl;
    std::cout << std::endl;

    if(argc != 2) {
        std::cout << "Usage (assuming an image demo.bmp in same folder):" << std::endl;
        std::cout << "sudo ./ejump_usb demo.bmp" << std::endl;
        return LIBTCON_SUCCESS;
    }

    /* Before we can communicate with the T-CON board, the interface
     * has to be initialized. The return value will be the number of
     * T-CON boards detected. If return value is 0 or less, an error
     * occured (in case of negative values, the error code's meaning
     * will be in accordance with the enum libtcon_error). */
    std::cout << "Trying to initialize T-CON board..." << std::endl;
    short tcon_num = tcon_init();
    if (tcon_num < 0) {
        std::cout << "Error: Failed to initialize T-CON board for reason: " << error_to_string[tcon_num] << std::endl;
        return tcon_num;
    } else if (tcon_num == 0) {
        std::cout << "Error: No T-CON boards detected!" << std::endl;
        return LIBTCON_ERROR_NO_DEVICE;
    } else {
        std::cout << "Successfully initialized T-CON board." << std::endl;
    }

    /* Defines the target T-CON board: Set this variable to the configured
     * ID of any connected T-CON board (e.g. by means of the rotary switch
     * on EJ8951-1W devices) to communicate with that specific board or
     * to -1 to communicate with the first T-CON board found. */
    short board_id = -1;

    /* Name of the image file to be displayed. */
    std::string img_filename{argv[1]};

    /* Waveform mode used for displaying the image. Refer to the
     * document "Information about E-Paper Waveform Modes" in the
     * downloads section of your eJump card for more information. */
    short wf_mode = 2;

    /* Query the width and height of the e-paper connect to the chosen
     * T-CON board. */
    short epd_width = tcon_get_panel_width(board_id);
    std::cout << "Width of e-paper: " << epd_width << "px" << std::endl;
    short epd_height = tcon_get_panel_height(board_id);
    std::cout << "Height of e-paper: " << epd_height << "px" << std::endl;

    /* Query the temperature of the T-CON board. */
    std::cout << "Querying board temperature..." << std::endl;
    short temperature = tcon_get_temp(board_id);
    if (temperature < 0){
        std::cout << "Error: Cannot read temperature for reason: " << error_to_string[temperature] << std::endl;
        return temperature;
    } else {
        std::cout << "Current temperature: " << temperature << "°C" << std::endl;
    }

    /* Clear the screen of the e-paper display. */
    std::cout << "Clearing the screen..." << std::endl;
    if (auto retval = tcon_clr_scr(board_id); retval != LIBTCON_SUCCESS) {
        std::cout << "Error: Could not clear screen for reason: " << error_to_string[retval] << std::endl;
        return retval;
    } else {
        std::cout << "Successfully cleared screen!" << std::endl;
    }

    /* Clear the image buffer of the T-CON board by setting it to a
     * well defined value (0xF0: white, 0x00: black). */
    std::cout << "Clearing image buffer of T-CON board..." << std::endl;
    if (auto retval = tcon_fill_img_all(board_id, 0x00); retval != LIBTCON_SUCCESS) {
        std::cout << "Unable to clear image buffer for reason: " << retval << std::endl;
        return retval;
    } else {
        std::cout << "Successfully cleared the image buffer!" << std::endl;
    }

    /* Now we need to load the image from file. We use OpenCV to do so.
     * However, we might use any third-party library to do so, as long
     * as the resulting image stream (i.e. a char array) has the same
     * format as the OpenCV image below. */
    std::cout << "Loading image from disk..." << std::endl;
    cv::Mat image = cv::imread(img_filename, cv::IMREAD_GRAYSCALE);
    cv::resize(image, image, cv::Size(epd_width, epd_height), 0, 0, cv::INTER_LINEAR);

    /* After loading the display into RAM, we transfer it to the T-CON
     * board. If you only want to update a part of the screen, you may
     * specify the area. */
    std::cout << "Transferring image into buffer of T-CON board..." << std::endl;
    if (auto retval = tcon_ld_img(board_id, image.data, epd_width, epd_height, 0, 0); retval != LIBTCON_SUCCESS) {
        std::cout << "Error: Could not transfer image for reason: " << error_to_string[retval] << std::endl;
        return retval;
    } else {
        std::cout << "Successfully transferred image!" << std::endl;
    }

    /* Now that the image is transfered, we can finally display it. */
    std::cout << "Displaying image onto the e-paper..." << std::endl;
    if (auto retval = tcon_dpy_img(board_id, wf_mode, epd_width, epd_height, 0, 0); retval != LIBTCON_SUCCESS) {
        std::cout << "Error: Could not display image for reason: " << error_to_string[retval] << std::endl;
        return retval;
    } else {
        std::cout << "Successfully displayed image!" << std::endl;
    }

    return LIBTCON_SUCCESS;
}
